--- normal_bios/rombios.org.c	2007-09-07 12:36:35.000000000 -0600
+++ bios/rombios.c	2007-09-14 10:51:01.000000000 -0600
@@ -124,8 +124,8 @@
 
 #define DEBUG_ROMBIOS      1
 
-#define DEBUG_ATA          0
-#define DEBUG_INT13_HD     0
+#define DEBUG_ATA          1
+#define DEBUG_INT13_HD     1
 #define DEBUG_INT13_CD     0
 #define DEBUG_INT13_ET     0
 #define DEBUG_INT13_FL     0
@@ -147,7 +147,8 @@
 #define LINUXBIOS	1
 
 #define BX_USE_ATADRV    1
-//#define BX_ELTORITO_BOOT 1
+#define BX_USE_FILO_ATADRV    1
+#define BX_ELTORITO_BOOT 1
 
 #define BX_MAX_ATA_INTERFACES   4
 #define BX_MAX_ATA_DEVICES      (BX_MAX_ATA_INTERFACES*2)
@@ -632,6 +633,9 @@
 #define BiosData ((bios_data_t  *) 0)
 
 #if BX_USE_ATADRV
+#if BX_USE_FILO_ATADRV
+	#include "filo_ide.h"
+#endif // BX_USE_FILO_ATADRV
   typedef struct {
     Bit16u heads;      // # heads
     Bit16u cylinders;  // # cylinders
@@ -732,6 +736,10 @@
 
     // ATA Driver data
     ata_t   ata;
+#if BX_USE_FILO_ATADRV
+    controller_t controllers[IDE_MAX_CONTROLLERS];
+    harddisk_info_t harddisk_info[IDE_MAX_DRIVES];
+#endif // BX_USE_FILO_ATADRV
 
 #if BX_ELTORITO_BOOT
     // El Torito Emulation data
@@ -780,7 +788,6 @@
     } dpt_t;
  
   #define Int13DPT ((dpt_t *) 0)
-
 #endif // BX_USE_ATADRV
 
 typedef struct {
@@ -909,6 +916,7 @@
 # if BX_USE_ATADRV
 
 // ATA / ATAPI driver
+void   mdelay();
 void   ata_init();
 void   ata_detect();
 void   ata_reset();
@@ -1952,6 +1960,10 @@
 
 #if BX_USE_ATADRV
 
+#if BX_USE_FILO_ATADRV
+	#include "filo_ide.c"
+#endif // BX_USE_FILO_ATADRV
+
 // ---------------------------------------------------------------------------
 // Start of ATA/ATAPI Driver
 // ---------------------------------------------------------------------------
@@ -2194,6 +2206,7 @@
 
     // Look for device
     outb(iobase1+ATA_CB_DH, slave ? ATA_CB_DH_DEV1 : ATA_CB_DH_DEV0);
+	mdelay(50);
     outb(iobase1+ATA_CB_SC, 0x55);
     outb(iobase1+ATA_CB_SN, 0xaa);
     outb(iobase1+ATA_CB_SC, 0xaa);
@@ -2206,6 +2219,7 @@
     sn = inb(iobase1+ATA_CB_SN);
 
     if ( (sc == 0x55) && (sn == 0xaa) ) {
+      BX_DEBUG_ATA("found a device #%d\n",device);
       write_byte(ebda_seg,&EbdaData->ata.devices[device].type,ATA_TYPE_UNKNOWN);
     
       // reset the channel
@@ -2215,6 +2229,7 @@
       outb(iobase1+ATA_CB_DH, slave ? ATA_CB_DH_DEV1 : ATA_CB_DH_DEV0);
       sc = inb(iobase1+ATA_CB_SC);
       sn = inb(iobase1+ATA_CB_SN);
+      BX_DEBUG_ATA("device %d detect sc = %02x sn= %02x\n",device,sc,sn);
       if ((sc==0x01) && (sn==0x01)) {
         cl = inb(iobase1+ATA_CB_CL);
         ch = inb(iobase1+ATA_CB_CH);
@@ -2231,9 +2246,10 @@
     }
 
     type=read_byte(ebda_seg,&EbdaData->ata.devices[device].type);
+    BX_DEBUG_ATA("device %d type = %x\n",device,type);
     
     // Now we send a IDENTIFY command to ATA device 
-    if(type == ATA_TYPE_ATA) {
+    if(type == ATA_TYPE_ATA || type == ATA_TYPE_UNKNOWN) {
       Bit32u sectors;
       Bit16u cylinders, heads, spt, blksize;
       Bit8u  translation, removable, mode;
@@ -2255,6 +2271,8 @@
 
       sectors   = read_dword(get_SS(),buffer+(60*2)); // word 60 and word 61
 
+    BX_DEBUG_ATA("device %d removable %d mode %d blksize %d cyl %d heads %d spt %d sectors %lx\n",device,removable,mode,blksize,cylinders,heads,spt,sectors);
+
       write_byte(ebda_seg,&EbdaData->ata.devices[device].device,ATA_DEVICE_HD);
       write_byte(ebda_seg,&EbdaData->ata.devices[device].removable, removable);
       write_byte(ebda_seg,&EbdaData->ata.devices[device].mode, mode);
@@ -2803,9 +2821,11 @@
   Bit16u ebda_seg=read_word(0x0040,0x000E);
   Bit16u iobase1, iobase2;
   Bit16u lcount, lbefore, lafter, count;
+  Bit16u loops;
   Bit8u  channel, slave;
   Bit8u  status, mode, lmode;
   Bit32u total, transfer;
+  Bit8u  err, sc;
 
   channel = device / 2;
   slave = device % 2;
@@ -2839,6 +2859,7 @@
   if (status & ATA_CB_STAT_BSY) return 2;
 
   outb(iobase2 + ATA_CB_DC, ATA_CB_DC_HD15 | ATA_CB_DC_NIEN);
+	/* Set up packet command */
   // outb(iobase1 + ATA_CB_FR, 0x00);
   // outb(iobase1 + ATA_CB_SC, 0x00);
   // outb(iobase1 + ATA_CB_SN, 0x00);
@@ -2846,10 +2867,27 @@
   outb(iobase1 + ATA_CB_CH, 0xfff0 >> 8);
   outb(iobase1 + ATA_CB_DH, slave ? ATA_CB_DH_DEV1 : ATA_CB_DH_DEV0);
   outb(iobase1 + ATA_CB_CMD, ATA_CMD_PACKET);
+	mdelay(1); /* spec says wait 400 ns , 1 ms is close enough */
+
+	/*Device should be busy */
+/*
+    status = inb(iobase1 + ATA_CB_STAT);
+    sc = inb(iobase1 + ATA_CB_SC);
+    err = inb(iobase1 + ATA_CB_ERR);
+	BX_DEBUG_ATA("Device should be busy and ready for data i.e. SC=01 STATUS=?8\n");
+	BX_DEBUG_ATA("ATA registers: SC %02x STATUS %02x ERR %02x\n",
+		sc,status,err);
+*/
 
   // Device should ok to receive command
   while (1) {
     status = inb(iobase1 + ATA_CB_STAT);
+/*
+    sc = inb(iobase1 + ATA_CB_SC);
+    err = inb(iobase1 + ATA_CB_ERR);
+	
+	BX_DEBUG_ATA("ATA registers2: SC %02x STATUS %02x ERR %02x\n",
+		sc,status,err);*/
     if ( !(status & ATA_CB_STAT_BSY) ) break;
     }
 
@@ -2888,11 +2926,31 @@
 
   if (inout == ATA_DATA_NO) {
     status = inb(iobase1 + ATA_CB_STAT);
+    sc = inb(iobase1 + ATA_CB_SC);
+    err = inb(iobase1 + ATA_CB_ERR);
+	BX_DEBUG_ATA("ATA_DATA_NO registers: SC %02x STATUS %02x ERR %02x\n",
+		sc,status,err);
     }
   else {
+	loops = 0;
   while (1) {
 
-      status = inb(iobase1 + ATA_CB_STAT);
+      status = inb(iobase1 + ATA_CB_STAT); // SPEC says read once and ignore
+	 mdelay(10);
+      //BX_DEBUG_ATA(".\n");
+      BX_DEBUG_ATA("ata_cmd_packet %d: waiting 1 status=%02x\n",loops,status);
+      //BX_DEBUG("%02x\n",status);
+
+	if (loops == 0) //first time through
+	{
+		do
+		{
+	 		mdelay(10); // This is mostly to see how long it takes
+      		status = inb(iobase1 + ATA_CB_STAT);
+      		BX_DEBUG_ATA("ata_cmd_packet %d: waiting 2, status=%02x\n",loops,status);
+			loops++;
+		} while (loops <20 && (status & ATA_CB_STAT_BSY));
+	}
 
       // Check if command completed
       if ( (status & (ATA_CB_STAT_BSY | ATA_CB_STAT_DRQ) ) ==0 ) break;
@@ -2902,11 +2960,11 @@
         return 3;
       }
 
-      // Device must be ready to send data
-      if ( (status & (ATA_CB_STAT_BSY | ATA_CB_STAT_RDY | ATA_CB_STAT_DRQ | ATA_CB_STAT_ERR) ) 
-            != (ATA_CB_STAT_RDY | ATA_CB_STAT_DRQ) ) {
-        BX_DEBUG_ATA("ata_cmd_packet : not ready (status %02x)\n", status);
-        return 4;
+      // Check for no error and Data Request
+      if ( (status & (ATA_CB_STAT_DRQ | ATA_CB_STAT_ERR) ) 
+            != (ATA_CB_STAT_DRQ) ) {
+        	BX_DEBUG_ATA("ata_cmd_packet : not ready 1 (status %02x)\n", status);
+        		//return 4;
         }
 
       // Normalize address
@@ -3049,7 +3107,7 @@
   // Final check, device must be ready
   if ( (status & (ATA_CB_STAT_BSY | ATA_CB_STAT_RDY | ATA_CB_STAT_DF | ATA_CB_STAT_DRQ | ATA_CB_STAT_ERR) ) 
          != ATA_CB_STAT_RDY ) {
-    BX_DEBUG_ATA("ata_cmd_packet : not ready (status %02x)\n", (unsigned) status);
+    BX_DEBUG_ATA("ata_cmd_packet : not ready 2 (status %02x)\n", status);
     return 4;
     }
 
@@ -3145,6 +3203,16 @@
 // ---------------------------------------------------------------------------
 
   void
+mdelay(msecs)
+Bit16u msecs;
+{
+  Bit16u i,j;
+	for (i=0;i<msecs;i++)
+		for (j=0;j<2048;j++)
+			inb(0x80);
+}
+
+  void
 cdemu_init()
 {
   Bit16u ebda_seg=read_word(0x0040,0x000E);
@@ -3191,6 +3259,9 @@
   // if not found
   if(device >= BX_MAX_ATA_DEVICES) return 2;
 
+  error = atapi_is_ready(device);
+  BX_DEBUG_ATA("atapi is ready returns %x\n",error);
+
   // Read the Boot Record Volume Descriptor
   memsetb(get_SS(),atacmd,0,12);
   atacmd[0]=0x28;                      // READ command
@@ -3201,7 +3272,10 @@
   atacmd[4]=(0x11 & 0x0000ff00) >> 8;
   atacmd[5]=(0x11 & 0x000000ff);
   if((error = ata_cmd_packet(device, 12, get_SS(), atacmd, 0, 2048L, ATA_DATA_IN, get_SS(), buffer)) != 0)
-    return 3;
+  {
+	bios_printf(BIOS_PRINTF_INFO | BIOS_PRINTF_SCREEN, "ata_cmd_packet returned %d\n",error);
+     return 3;
+  }
 
   // Validity checks
   if(buffer[0]!=0)return 4;
@@ -9795,9 +9869,16 @@
   ;;
   ;; ATA/ATAPI driver setup
   ;;
+#if BX_USE_FILO_ATADRV
+  ;;
+  call _ide_probe_all
+  ;;
+#else
+  ;;
   call _ata_init
   call _ata_detect
   ;;
+#endif // BX_USE_FILO_ATADRV
 #else // BX_USE_ATADRV
 
   ;;
